#include <stdio.h>
#include <assert.h>
#include "SDL.h"
#include "bitfont.h"

/**************************************************************************/

int InitBitFont(struct BitFont *font)
{
  memset(font,0,sizeof(font));
}

void KillBitFont(struct BitFont *font)
{
  if(font->Data)
  {
    free(font->Data);
    font->Data=0;
  }

  if(font->Letters)
  {
    free(font->Letters);
    font->Letters=0;
  }
}

int LoadBitFontFile(struct BitFont *font,char *filename)
{
  int size;
  unsigned int *data;
  FILE *file;
  int result;

  file=fopen(filename,"rb");
  if(file==0)
    return 0;

  fseek(file,0,SEEK_END);
  size=ftell(file);
  fseek(file,0,SEEK_SET);

  if(size<12)
  {
    fclose(file);
    return 0;
  }

  data=(unsigned int *)malloc(size);
  if(data==0)
  {
    fclose(file);
    return 0;
  }

  if(fread(data,1,size,file)!=size)
  {
    fclose(file);
    free(data);
    return 0;
  }

  fclose(file);

  result=LoadBitFontMemory(font,data,size);
  
  free(data);
  return result;
}

int LoadBitFontMemory(struct BitFont *font,unsigned int *mem,int size)
{
  int pos;
  int xs,v;

  InitBitFont(font);
  pos=0;

  if(mem[pos++]!='F'+('O'<<8)+('N'<<16)+('T'<<24))
    return 0;
 
  v=mem[pos++];
  xs=mem[pos++];
  font->Height=mem[pos++];
  font->First=mem[pos++];
  font->Count=mem[pos++];
  font->Bad=mem[pos++];

  if((xs&31)!=0) return 0;
  font->BPR=xs/8;

  font->Letters = (struct BitFontLetter *)
    malloc(sizeof(struct BitFontLetter) * font->Count);
  font->Data = (unsigned char *)
    malloc(font->Height * font->BPR);

  if(font->Letters==0 || font->Data==0)
  {
    KillBitFont(font);
    return 0;
  }

  memcpy(font->Letters,mem+pos,sizeof(struct BitFontLetter)*font->Count);
  pos+=sizeof(struct BitFontLetter)*font->Count/4;

  memcpy(font->Data,mem+pos,font->Height*font->BPR);
  pos+=(font->Height*font->BPR)/4;

  if(mem[pos++]!='f'+('o'<<8)+('n'<<16)+('t'<<24))
  {
    KillBitFont(font);
    return 0; 
  }
  if(pos*4 != size)
  {
    KillBitFont(font);
    return 0;
  }

  return 1;
}

int SaveBitFontFile(struct BitFont *font,char *filename)
{
  int size;
  unsigned int *data;
  FILE *file;  

  if(SaveBitFontMemory(font,&data,&size)==0)
    return 0;

  file=fopen(filename,"wb");
  if(file==0)
  {
    free(data);
    return 0;
  }

  if(fwrite(data,1,size,file)!=size)
  {
    fclose(file);
    free(data);
    return 0;
  }

  fclose(file);
  free(data);

  return 1;
}

int SaveBitFontMemory(struct BitFont *font,unsigned int **pm,int *ps)
{
  unsigned int *data;
  int size;
  int pos;
 
  *pm=0;
  *ps=0;
  size = 8*4 + font->Count*sizeof(struct BitFontLetter);
  size += font->BPR * font->Height;   

  if((font->BPR&3)!=0)       /* alignment restriction 32 pixels */
    return;

  data = (unsigned int *)malloc(size);
  if(!data)
    return 0;

  pos=0;
  data[pos++]='F'+('O'<<8)+('N'<<16)+('T'<<24);
  data[pos++]=0;
  data[pos++]=font->BPR*8;
  data[pos++]=font->Height;
  data[pos++]=font->First;
  data[pos++]=font->Count;
  data[pos++]=font->Bad;

  memcpy(data+pos,font->Letters,sizeof(struct BitFontLetter)*font->Count);
  pos+=(sizeof(struct BitFontLetter)*font->Count)/4;
  memcpy(data+pos,font->Data,font->BPR*font->Height);
  pos+=(font->BPR*font->Height)/4;

  data[pos++]='f'+('o'<<8)+('n'<<16)+('t'<<24);
  
  assert(size=pos*4);

  *ps=size;
  *pm=data;
  
  return 1;
}

/**************************************************************************/

static void Blit(unsigned char *dp,int dbpr,int dx,int dy,
                 unsigned char *sp,int sbpr,int sx,int sy,
                 int xl,int yl,unsigned int col)
{
  unsigned char *s,*ss;
  unsigned char *d,*dd;
  unsigned char byte;
  unsigned int mask;
  int pre,post,middle;
  int y,i; 

  if(xl==0 || yl==0)
    return;

  s=sp+sy*sbpr+sx/8;
  d=dp+dy*dbpr+dx;

  pre=sx&7;
  post=(sx+xl-1)&7;
  middle=(sx+xl-1)/8 - sx/8 + 1;

  if(pre!=0)
    middle--;
  if(post!=7)
    middle--;

  if(middle<0)
  {
    for(y=0;y<yl;y++)
    {
      dd=d;
      mask=(((1<<(8-pre))-1) & (0x7f80>>post));
      byte=*s;
      if(mask&0x80) { if(byte&0x80) *dd=col; dd++; }
      if(mask&0x40) { if(byte&0x40) *dd=col; dd++; }
      if(mask&0x20) { if(byte&0x20) *dd=col; dd++; }
      if(mask&0x10) { if(byte&0x10) *dd=col; dd++; }
      if(mask&0x08) { if(byte&0x08) *dd=col; dd++; }
      if(mask&0x04) { if(byte&0x04) *dd=col; dd++; }
      if(mask&0x02) { if(byte&0x02) *dd=col; dd++; }
      if(mask&0x01) { if(byte&0x01) *dd=col; dd++; }
    
      d+=dbpr;
      s+=sbpr;
    }
  }
  else
  {
    for(y=0;y<yl;y++)
    {
      ss=s;
      dd=d;
      if(pre!=0)
      {
        byte=*ss++;
        if(pre<=1) { if(byte&0x40) *dd=col; dd++; }
        if(pre<=2) { if(byte&0x20) *dd=col; dd++; }
        if(pre<=3) { if(byte&0x10) *dd=col; dd++; }
        if(pre<=4) { if(byte&0x08) *dd=col; dd++; }
        if(pre<=5) { if(byte&0x04) *dd=col; dd++; }
        if(pre<=6) { if(byte&0x02) *dd=col; dd++; }
                  { if(byte&0x01) *dd=col; dd++; }
      }

      for(i=0;i<middle;i++)
      {
        byte=*ss++;
        if(byte&0x80) *dd=col; dd++;
        if(byte&0x40) *dd=col; dd++;
        if(byte&0x20) *dd=col; dd++;
        if(byte&0x10) *dd=col; dd++;
        if(byte&0x08) *dd=col; dd++;
        if(byte&0x04) *dd=col; dd++;
        if(byte&0x02) *dd=col; dd++;
        if(byte&0x01) *dd=col; dd++;
      }

      if(post!=7)
      {
        byte=*ss++;
                    { if(byte&0x80) *dd=col; dd++; }
        if(post>=1) { if(byte&0x40) *dd=col; dd++; }
        if(post>=2) { if(byte&0x20) *dd=col; dd++; }
        if(post>=3) { if(byte&0x10) *dd=col; dd++; }
        if(post>=4) { if(byte&0x08) *dd=col; dd++; }
        if(post>=5) { if(byte&0x04) *dd=col; dd++; }
        if(post>=6) { if(byte&0x02) *dd=col; dd++; }
      }

      d+=dbpr;
      s+=sbpr;
    }
  }
}

/* This is just a copy of Blit() with the destination data type changed */
static void Blit16(Uint16 *dp,int dbpr,int dx,int dy,
                 unsigned char *sp,int sbpr,int sx,int sy,
                 int xl,int yl,unsigned int col)
{
  unsigned char *s,*ss;
  Uint16 *d,*dd;
  unsigned char byte;
  unsigned int mask;
  int pre,post,middle;
  int y,i; 

  if(xl==0 || yl==0)
    return;

  dbpr /= 2;
  s=sp+sy*sbpr+sx/8;
  d=dp+dy*dbpr+dx;

  pre=sx&7;
  post=(sx+xl-1)&7;
  middle=(sx+xl-1)/8 - sx/8 + 1;

  if(pre!=0)
    middle--;
  if(post!=7)
    middle--;

  if(middle<0)
  {
    for(y=0;y<yl;y++)
    {
      dd=d;
      mask=(((1<<(8-pre))-1) & (0x7f80>>post));
      byte=*s;
      if(mask&0x80) { if(byte&0x80) *dd=col; dd++; }
      if(mask&0x40) { if(byte&0x40) *dd=col; dd++; }
      if(mask&0x20) { if(byte&0x20) *dd=col; dd++; }
      if(mask&0x10) { if(byte&0x10) *dd=col; dd++; }
      if(mask&0x08) { if(byte&0x08) *dd=col; dd++; }
      if(mask&0x04) { if(byte&0x04) *dd=col; dd++; }
      if(mask&0x02) { if(byte&0x02) *dd=col; dd++; }
      if(mask&0x01) { if(byte&0x01) *dd=col; dd++; }
    
      d+=dbpr;
      s+=sbpr;
    }
  }
  else
  {
    for(y=0;y<yl;y++)
    {
      ss=s;
      dd=d;
      if(pre!=0)
      {
        byte=*ss++;
        if(pre<=1) { if(byte&0x40) *dd=col; dd++; }
        if(pre<=2) { if(byte&0x20) *dd=col; dd++; }
        if(pre<=3) { if(byte&0x10) *dd=col; dd++; }
        if(pre<=4) { if(byte&0x08) *dd=col; dd++; }
        if(pre<=5) { if(byte&0x04) *dd=col; dd++; }
        if(pre<=6) { if(byte&0x02) *dd=col; dd++; }
                  { if(byte&0x01) *dd=col; dd++; }
      }

      for(i=0;i<middle;i++)
      {
        byte=*ss++;
        if(byte&0x80) *dd=col; dd++;
        if(byte&0x40) *dd=col; dd++;
        if(byte&0x20) *dd=col; dd++;
        if(byte&0x10) *dd=col; dd++;
        if(byte&0x08) *dd=col; dd++;
        if(byte&0x04) *dd=col; dd++;
        if(byte&0x02) *dd=col; dd++;
        if(byte&0x01) *dd=col; dd++;
      }

      if(post!=7)
      {
        byte=*ss++;
                    { if(byte&0x80) *dd=col; dd++; }
        if(post>=1) { if(byte&0x40) *dd=col; dd++; }
        if(post>=2) { if(byte&0x20) *dd=col; dd++; }
        if(post>=3) { if(byte&0x10) *dd=col; dd++; }
        if(post>=4) { if(byte&0x08) *dd=col; dd++; }
        if(post>=5) { if(byte&0x04) *dd=col; dd++; }
        if(post>=6) { if(byte&0x02) *dd=col; dd++; }
      }

      d+=dbpr;
      s+=sbpr;
    }
  }
}

/**************************************************************************/

void Text(SDL_Surface *dest,struct BitFont *font,int x,int y,
  unsigned int col,char *text,int size)
{
  int i,c;

  if(size==-1)
  {
    size=0;
    while(text[size]) size++;
  }

  for(i=0;i<size;i++)
  {
    c=text[i];
    c-=font->First;
    if(c<0 || c>font->Count)
      c=font->Bad-font->First;

    x+=font->Letters[c].prekern;

    switch (dest->format->BytesPerPixel) {
      case 1:
        Blit((Uint8 *)dest->pixels,dest->pitch,x,y,
             font->Data,font->BPR,font->Letters[c].location,0,
             font->Letters[c].cellwidth,font->Height,col);
        break;
      case 2:
        Blit16((Uint16 *)dest->pixels,dest->pitch,x,y,
             font->Data,font->BPR,font->Letters[c].location,0,
             font->Letters[c].cellwidth,font->Height,col);
        break;
      default:
        /* Blit not supported... */
        break;
    }

    x+=font->Letters[c].totalwidth;
  }
}

void TextRect(SDL_Surface *dest,struct BitFont *font,
  int x1,int y1,int x2,int y2,int align,
  unsigned int col,char *text,int size)
{
  int xs,ys;
  int x,y;

  if(size==-1)
  {
    size=0;
    while(text[size]) size++;
  }

  xs=TextWidth(font,text,size);
  ys=font->Height;

  if(align & POS_LEFT)
    x=x1;
  else if(align & POS_RIGHT)
    x=x2-xs;
  else
    x=(x1+x2-xs)/2;

  if(align & POS_TOP)
    y=y1;
  else if(align & POS_BOTTOM)
    y=y2-ys;
  else
    y=(y1+y2-ys)/2;

  Text(dest,font,x,y,col,text,size);
}

int TextWidth(struct BitFont *font,char *text,int size)
{
  int i;
  int width;
  int c;

  if(size==-1)
  {
    size=0;
    while(text[size]) size++;
  }

  width=0;
  for(i=0;i<size;i++)
  {
    c = text[i];
    c -= font->First;
    if(c<0 || c>font->Count)
      c = font->Bad - font->First;
  
    width += font->Letters[c].prekern + font->Letters[c].totalwidth;
  }

  return width;
}

/**************************************************************************/
